package top.ninwoo.cloudcenter.service.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;
import top.ninwoo.cloudcenter.entity.SeparatedClusterConfig;
import top.ninwoo.cloudcenter.register.CloudRegisterCenter;
import top.ninwoo.cloudcenter.service.CloudService;
import top.ninwoo.common.entity.ClusterConfig;
import top.ninwoo.common.entity.ContainerDescription;
import top.ninwoo.common.entity.NetworkTopology;

import javax.annotation.Resource;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

@Service
public class CloudServiceImpl implements CloudService {
    private static final Logger LOG = LoggerFactory.getLogger(CloudServiceImpl.class);
    private static final Random randomInt = new Random(14);
    private static final String CREATE_CLUSTER = "/createCluster";
    private static final String DELETE_CLUSTER = "/delCluster";
    private static final String DROP_DOCKER_NETWORK = "dropDockerNetwork";
    private static final String CANCEL_DROP_DOCKER_NETWORK = "cancdelDropDockerNetwork";
    // 全部的逻辑拓扑
    // 全部的逻辑拓扑
    private ConcurrentHashMap<Long, NetworkTopology> allLogicalTopo = new ConcurrentHashMap<>();
    // 用于存储配置文件
    private ConcurrentHashMap<Long, List<SeparatedClusterConfig>> allClusterConfig
            = new ConcurrentHashMap<>();

    private ConcurrentHashMap<Long, Map<String, Set<String>>> allAppStatus
            = new ConcurrentHashMap<>();

    @Resource
    private CloudRegisterCenter cloudRegisterCenter;

    @Resource
    private RestTemplate restTemplate;


    /**
     * 这里的拓扑必须限定为逻辑拓扑
     * @param clusterConfig
     * @return
     */
    @Override
    public String initCluster(ClusterConfig clusterConfig) {

        // TODO: step1: 校验配置是否合法
        check(clusterConfig);

        // 在数据库中注册该节点
        registerClusterInDataBase(clusterConfig);
        // TODO: step2: 拆分配置

        // TODO: step3: 下发配置到不同的边缘服务节点
        // TODO: 保存配置和边缘节点的映射关系
        return "";
    }

    private ClusterConfig registerClusterInDataBase(ClusterConfig clusterConfig) {
        clusterConfig.setId(11111);
        return clusterConfig;
    }

    private boolean check(ClusterConfig clusterConfig) {
        return true;
    }

    private List<SeparatedClusterConfig> separateClusterConfig(ClusterConfig clusterConfig) {
        // 先校验clusterConfig的模式
        String type = clusterConfig.getType();
        if("single".equals(type)) {
            return singleSeparateClusterConfig(clusterConfig);
        }
        if("random".equals(type)) {
            return randomSeparateClusterConfig(clusterConfig);
        }
        if("all".equals(type)) {
            return allSeparatedClusterConfig(clusterConfig);
        }
        if(type.startsWith("size:")) {
            int size = Integer.parseInt(type.split(":")[1]);
            return fixedSpeparatedClusterConfig(clusterConfig, size);
        }
        throw new IllegalArgumentException("clusterConfig's " + type + " is Illegal!");
    }

    private List<SeparatedClusterConfig> fixedSpeparatedClusterConfig(ClusterConfig clusterConfig, int size) {

        List<String> nodes = chooseEdgeNodes(size);

        Map<String, SeparatedClusterConfig> separatedClusterConfigs = new HashMap<>();

        for (int i = 0; i < nodes.size(); i++) {
            SeparatedClusterConfig separatedClusterConfig = new SeparatedClusterConfig();
            separatedClusterConfig.setEdgeNodeId(nodes.get(i));
            ClusterConfig config = new ClusterConfig();
            config.setId(clusterConfig.getId());
            // 只有这里不同
            config.setDockers(new ArrayList<>());
            NetworkTopology networkTopology = new NetworkTopology();
            config.setTopology(networkTopology);
            config.setOwner(clusterConfig.getOwner());
            config.setCreateTime(clusterConfig.getCreateTime());
            config.setType(clusterConfig.getType());
            separatedClusterConfig.setClusterConfig(config);
            separatedClusterConfigs.put(nodes.get(i), separatedClusterConfig);

        }


        int index = 0;
        // TODO: 拆分配置,构建出新的配置
        List<ContainerDescription> dockers = clusterConfig.getDockers();
        for (ContainerDescription docker : dockers) {
            if("one".equals(docker.getMode())) {
                // 该docker全部部署到同一个集群上
                // TODO: 这里可以添加一些选择标准
                SeparatedClusterConfig separatedClusterConfig = separatedClusterConfigs.get(nodes.get(index));
                ClusterConfig edgeClusterConfig = separatedClusterConfig.getClusterConfig();
                // 更新edgeColusterClusterConfig上的配置
                edgeClusterConfig.getDockers().add(docker);
                NetworkTopology topology = edgeClusterConfig.getTopology();
                String[] appNames = topology.getAppNames();
                int[][] topo = topology.getTopology();
                // TODO: 扩大规模
                if(appNames == null) {
                    appNames = new String[]{docker.getDockerContainer().getName()};
                    topo = new int[][]{{0}};
                } else {
                    appNames = Arrays.copyOf(appNames, appNames.length + 1);
                    appNames[appNames.length - 1] = docker.getDockerContainer().getName();
                    // 可能出问题
                    topo = new int[topo.length + 1][topo.length + 1];
                }
                topology.setAppNames(appNames);
                topology.setTopology(topo);
            }
            else if(docker.getMode().equals("roundbin")) {
                int replicas = docker.getReplicas();
                int average = replicas / nodes.size();
                for (int i = 0; i < nodes.size(); i++) {
                    // TODO: 更新集群，更新
                }
            }

            index = (index + 1) % nodes.size();
        }
        return null;
    }

    private List<SeparatedClusterConfig> allSeparatedClusterConfig(ClusterConfig clusterConfig) {
        List<String> availableEdgeNodes = cloudRegisterCenter.getAvailableEdgeNodes();
        return null;
    }

    private List<SeparatedClusterConfig> randomSeparateClusterConfig(ClusterConfig clusterConfig) {
        // 创建随机的
        return null;
    }

    private List<SeparatedClusterConfig> singleSeparateClusterConfig(ClusterConfig clusterConfig) {
        // 选择一个节点
        List<String> nodes = chooseEdgeNodes(1);
        return null;
    }

    private List<String> chooseEdgeNodes(int num) {
        // TODO 可以扩展更多选择方案，比如已计算优先，或者内存优先

        ArrayList<String> res = new ArrayList<>(num);

        // step1: 获取全部可用的边缘节点
        List<String> availableEdgeNodes = cloudRegisterCenter.getAvailableEdgeNodes();

        if(availableEdgeNodes.size() <= num) {
            return availableEdgeNodes;
        }

        // step2: 随机挑选一个节点
        for (int i = 0; i < num; i++) {
            int index = randomInt.nextInt(availableEdgeNodes.size());
            res.add(availableEdgeNodes.get(index));
            availableEdgeNodes.remove(index);
        }
        return res;
    }

    /**
     * 下发clusterConfigs到边缘节点
     * @param clusterConfigs
     * @return
     */
    @Override
    public List<SeparatedClusterConfig> sendClusterConfigToEdgeNode(List<SeparatedClusterConfig> clusterConfigs) {
        if(clusterConfigs.size() == 0) {
            LOG.warn("下发的集群配置为空。");
            return clusterConfigs;
        }
        Set<String> allAppNames = new HashSet<>();
        // 这里应该做一个判断，edgeNode的ip是否可用，如果不可用，或者未设定，将随机挑选一个进行设置
        clusterConfigs.forEach(c -> {
            if (c.getEdgeNodeId() == null || "".equals(c.getEdgeNodeId())) {
                String edgeNode = chooseEdgeNodes(1).get(0);
                c.setEdgeNodeId(edgeNode);
            }
            // 下发配置
            ResponseEntity<Long> response = restTemplate.postForEntity("http://" + c.getEdgeNodeId() + CREATE_CLUSTER, c.getClusterConfig(), Long.class);
            if(!response.getStatusCode().is2xxSuccessful()) {
                LOG.error("集群配置下发失败:{}:{}", c.getEdgeNodeId(), response.getBody());

            }

            // 构建初始化的逻辑拓扑
            for (String appName : c.getClusterConfig().getTopology().getAppNames()) {
                if(!appName.startsWith("br")) {
                    allAppNames.add(appName);
                    if (!allAppStatus.containsKey(c.getClusterConfig().getId())) {
                        allAppStatus.put(c.getClusterConfig().getId(), new HashMap<>());
                    }
                    // 保存appnames
                    Map<String, Set<String>> clusterAppConfig = allAppStatus.get(c.getClusterConfig().getId());
                    if(!clusterAppConfig.containsKey(appName)) {
                        clusterAppConfig.put(appName, new HashSet<>());
                    }
                    Set<String> edgeNodes = clusterAppConfig.get(appName);
                    edgeNodes.add(c.getEdgeNodeId());
                }
            }
            if(!allClusterConfig.containsKey(c.getClusterConfig().getId())) {
                allClusterConfig.put(c.getClusterConfig().getId(), new ArrayList<>());
            }
            allClusterConfig.get(c.getClusterConfig().getId()).add(c);
        });

        allLogicalTopo.put(clusterConfigs.get(0).getClusterConfig().getId(), initNetworkTopology(allAppNames));

        return clusterConfigs;
    }

    private NetworkTopology initNetworkTopology(Set<String> allAppNames) {
        NetworkTopology networkTopology = new NetworkTopology();
        networkTopology.setTopologyId(1);
        networkTopology.setAppNames(allAppNames.toArray(new String[allAppNames.size()]));
        int[][] topo = new int[allAppNames.size()][allAppNames.size()];
        for (int i = 0; i < allAppNames.size(); i++) {
            for (int j = 0; j < allAppNames.size(); j++) {
                topo[i][j] = 1;
            }
        }
        networkTopology.setTopology(topo);
        return networkTopology;
    }

    /**
     * 删除全部的容器集群
     * @param clusterId
     * @return
     */
    @Override
    public boolean deleteClusterFromEdgeNode(long clusterId) {
        Map<String, Object> parameters = new HashMap<>();
        parameters.put("clusterId", clusterId);
        List<SeparatedClusterConfig> separatedClusterConfigs = allClusterConfig.get(clusterId);
        separatedClusterConfigs.forEach(c -> {
            // 调用远程借口删除服务
            restTemplate.getForEntity("http://" + c.getEdgeNodeId() + DELETE_CLUSTER + "?clusterId=" + clusterId, String.class);
        });
        return true;
    }

    /**
     * 获取cluster的全网拓扑
     * @param clusterId
     * @return
     */
    public NetworkTopology getClusterTopology(long clusterId) {
        return null;
    }

    /**
     * 设置网络的逻辑拓扑
     * @param clusterId
     * @param topology
     */
    public void updateLogicalTopology(long clusterId, NetworkTopology topology) {
        // 校验下topo是否合法
        for (String appName : topology.getAppNames()) {
            if(appName.contains("br")) {
                throw new RuntimeException("暂时不支持非逻辑网络拓扑！");
            }
        }
        // 更新topo配置
        // 获取原来的网络topo
        if (!allLogicalTopo.containsKey(clusterId)) {
            throw new RuntimeException("不存在的topo，无法进行更新");
        }

        NetworkTopology networkTopology = allLogicalTopo.get(clusterId);
        String[] appNames = networkTopology.getAppNames();
        if(appNames.length != topology.getAppNames().length) {
            throw new RuntimeException("更新的topo不一致");
        }
        for (int i = 0; i < appNames.length; i++) {
            if (appNames[i] != topology.getAppNames()[i]) {
                throw new RuntimeException("更新的名称不一致");
            }
        }

        int[][] oldTopo = networkTopology.getTopology();
        int[][] newTopo = topology.getTopology();

        // 对比更新
        for (int i = 0; i < appNames.length; i++) {
            for (int j = 0; j < appNames.length; j++) {
                if(oldTopo[i][j] == 1 && newTopo[i][j] == 0) {
                    // 阻断连接
                    // TODO step1 找到appNames对应的边缘服务器
                    Set<String> edgeNodesIdSrc = allAppStatus.get(clusterId).get(appNames[i]);
                    // TODO step2 更新网络
                    Set<String> edgeNodesIdDst = allAppStatus.get(clusterId).get(appNames[j]);
                    // 获取全部的Dst 容器网络
                    // step2 调用远程接口，添加阻断连接
                }
                else if(oldTopo[i][j] == 0 && newTopo[i][j] == 1) {
                    // 创建连接
                    // step1 找到appNames对应的边缘服务器
                    // step2 调用远程接口，删除阻断连接
                }
            }
        }
    }

}
